import logging
import click
import aiohttp

from swh.vault.api.server import make_app_from_configfile


CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('--config-file', '-C', default=None,
              type=click.Path(exists=True, dir_okay=False,),
              help="Configuration file.")
@click.option('--log-level', '-l', default='INFO',
              type=click.Choice(logging._nameToLevel.keys()),
              help="Log level (default to INFO)")
@click.option('--no-stdout', is_flag=True, default=False,
              help="Do NOT output logs on the console")
@click.option('--host', default='0.0.0.0', help="Host to run the server")
@click.option('--port', default=5005, type=click.INT,
              help="Binding port of the server")
@click.option('--debug/--nodebug', default=True,
              help="Indicates if the server should run in debug mode")
@click.pass_context
def cli(ctx, config_file, log_level, no_stdout, host, port, debug):
    """Software Heritage Vault API server

    """
    from swh.scheduler.celery_backend.config import setup_log_handler
    log_level = setup_log_handler(
        loglevel=log_level, colorize=False,
        format='[%(levelname)s] %(name)s -- %(message)s',
        log_console=not no_stdout)

    try:
        app = make_app_from_configfile(config_file, debug=debug)
    except EnvironmentError as e:
        click.echo(e.msg, err=True)
        ctx.exit(1)

    aiohttp.web.run_app(app, host=host, port=int(port))


def main():
    return cli(auto_envvar_prefix='SWH_VAULT')


if __name__ == '__main__':
    main()



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information
import logging

logger = logging.getLogger(__name__)


def get_vault(cls='remote', args={}):
    """
    Get a vault object of class `vault_class` with arguments
    `vault_args`.

    Args:
        vault (dict): dictionary with keys:
        - cls (str): vault's class, either 'remote'
        - args (dict): dictionary with keys

    Returns:
        an instance of VaultBackend (either local or remote)

    Raises:
        ValueError if passed an unknown storage class.

    """
    if cls == 'remote':
        from .api.client import RemoteVaultClient as Vault
    elif cls == 'local':
        from swh.scheduler import get_scheduler
        from swh.storage import get_storage
        from swh.vault.cache import VaultCache
        from swh.vault.backend import VaultBackend as Vault
        args['cache'] = VaultCache(**args['cache'])
        args['storage'] = get_storage(**args['storage'])
        args['scheduler'] = get_scheduler(**args['scheduler'])
    else:
        raise ValueError('Unknown storage class `%s`' % cls)
    logger.debug('Instantiating %s with %s' % (Vault, args))
    return Vault(**args)



import pytest
import glob
import os
import pkg_resources.extern.packaging.version

from swh.core.utils import numfile_sortkey as sortkey
from swh.vault import get_vault
from swh.vault.tests import SQL_DIR
from swh.storage.tests import SQL_DIR as STORAGE_SQL_DIR
from pytest_postgresql import factories


pytest_v = pkg_resources.get_distribution("pytest").parsed_version
if pytest_v < pkg_resources.extern.packaging.version.parse('3.9'):
    @pytest.fixture
    def tmp_path(request):
        import tempfile
        import pathlib
        with tempfile.TemporaryDirectory() as tmpdir:
            yield pathlib.Path(tmpdir)


def db_url(name, postgresql_proc):
    return 'postgresql://{user}@{host}:{port}/{dbname}'.format(
        host=postgresql_proc.host,
        port=postgresql_proc.port,
        user='postgres',
        dbname=name)


postgresql2 = factories.postgresql('postgresql_proc', 'tests2')


@pytest.fixture
def swh_vault(request, postgresql_proc, postgresql, postgresql2, tmp_path):

    for sql_dir, pg in ((SQL_DIR, postgresql), (STORAGE_SQL_DIR, postgresql2)):
        dump_files = os.path.join(sql_dir, '*.sql')
        all_dump_files = sorted(glob.glob(dump_files), key=sortkey)

        cursor = pg.cursor()
        for fname in all_dump_files:
            with open(fname) as fobj:
                # disable concurrent index creation since we run in a
                # transaction
                cursor.execute(fobj.read().replace('concurrently', ''))
        pg.commit()

    vault_config = {
        'db': db_url('tests', postgresql_proc),
        'storage': {
            'cls': 'local',
            'args': {
                'db': db_url('tests2', postgresql_proc),
                'objstorage': {
                    'cls': 'pathslicing',
                    'args': {
                        'root': str(tmp_path),
                        'slicing': '0:1/1:5',
                    },
                },
            },
        },
        'cache': {
            'cls': 'pathslicing',
            'args': {
                'root': str(tmp_path),
                'slicing': '0:1/1:5',
                'allow_delete': True,
            }
        },
        'scheduler': {
            'cls': 'remote',
            'args': {
                'url': 'http://swh-scheduler:5008',
            },
        },
    }

    return get_vault('local', vault_config)



# Copyright (C) 2017  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information
import os

from swh.core.config import load_named_config, read as read_config
from swh.storage import get_storage
from swh.vault import get_vault
from swh.vault.cookers.base import DEFAULT_CONFIG_PATH, DEFAULT_CONFIG
from swh.vault.cookers.directory import DirectoryCooker
from swh.vault.cookers.revision_flat import RevisionFlatCooker
from swh.vault.cookers.revision_gitfast import RevisionGitfastCooker

COOKER_TYPES = {
    'directory': DirectoryCooker,
    'revision_flat': RevisionFlatCooker,
    'revision_gitfast': RevisionGitfastCooker,
}


def get_cooker_cls(obj_type):
    return COOKER_TYPES[obj_type]


def get_cooker(obj_type, obj_id):
    if 'SWH_CONFIG_FILENAME' in os.environ:
        cfg = read_config(os.environ['SWH_CONFIG_FILENAME'], DEFAULT_CONFIG)
    else:
        cfg = load_named_config(DEFAULT_CONFIG_PATH, DEFAULT_CONFIG)
    cooker_cls = get_cooker_cls(obj_type)
    if 'vault' not in cfg:
        raise ValueError("missing '%vault' configuration")

    vcfg = cfg['vault']
    if vcfg['cls'] != 'remote':
        raise EnvironmentError(
            "This vault backend can only be a 'remote' "
            "configuration", err=True)
    args = vcfg['args']
    if 'storage' not in args:
        args['storage'] = cfg.get('storage')

    if not args.get('storage'):
        raise ValueError(
            "invalid configuration; missing 'storage' config entry.")

    storage = get_storage(**args.pop('storage'))
    backend = get_vault(**vcfg)

    return cooker_cls(obj_type, obj_id,
                      backend=backend, storage=storage,
                      max_bundle_size=cfg['max_bundle_size'])



# Copyright (C) 2016-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
import aiohttp.web
import asyncio
import collections

from swh.core import config
from swh.core.api_async import (SWHRemoteAPI,
                                encode_data_server as encode_data,
                                decode_request)
from swh.model import hashutil
from swh.vault import get_vault
from swh.vault.cookers import COOKER_TYPES
from swh.vault.backend import NotFoundExc


DEFAULT_CONFIG_PATH = 'vault/server'
DEFAULT_CONFIG = {
    'storage': ('dict', {
        'cls': 'remote',
        'args': {
            'url': 'http://localhost:5002/',
        },
    }),
    'cache': ('dict', {
        'cls': 'pathslicing',
        'args': {
            'root': '/srv/softwareheritage/vault',
            'slicing': '0:1/1:5',
        },
    }),
    'client_max_size': ('int', 1024 ** 3),
    'vault': ('dict', {
        'cls': 'local',
        'args': {
            'db': 'dbname=softwareheritage-vault-dev',
        },
    }),
    'scheduler': ('dict', {
        'cls': 'remote',
        'args': {
            'url': 'http://localhost:5008/',
        }
    }),
}


@asyncio.coroutine
def index(request):
    return aiohttp.web.Response(body="SWH Vault API server")


# Web API endpoints

@asyncio.coroutine
def vault_fetch(request):
    obj_type = request.match_info['type']
    obj_id = request.match_info['id']

    if not request.app['backend'].is_available(obj_type, obj_id):
        raise aiohttp.web.HTTPNotFound

    return encode_data(request.app['backend'].fetch(obj_type, obj_id))


def user_info(task_info):
    return {'id': task_info['id'],
            'status': task_info['task_status'],
            'progress_message': task_info['progress_msg'],
            'obj_type': task_info['type'],
            'obj_id': hashutil.hash_to_hex(task_info['object_id'])}


@asyncio.coroutine
def vault_cook(request):
    obj_type = request.match_info['type']
    obj_id = request.match_info['id']
    email = request.query.get('email')
    sticky = request.query.get('sticky') in ('true', '1')

    if obj_type not in COOKER_TYPES:
        raise aiohttp.web.HTTPNotFound

    try:
        info = request.app['backend'].cook_request(obj_type, obj_id,
                                                   email=email, sticky=sticky)
    except NotFoundExc:
        raise aiohttp.web.HTTPNotFound

    # TODO: return 201 status (Created) once the api supports it
    return encode_data(user_info(info))


@asyncio.coroutine
def vault_progress(request):
    obj_type = request.match_info['type']
    obj_id = request.match_info['id']

    info = request.app['backend'].task_info(obj_type, obj_id)
    if not info:
        raise aiohttp.web.HTTPNotFound

    return encode_data(user_info(info))


# Cookers endpoints

@asyncio.coroutine
def set_progress(request):
    obj_type = request.match_info['type']
    obj_id = request.match_info['id']
    progress = yield from decode_request(request)
    request.app['backend'].set_progress(obj_type, obj_id, progress)
    return encode_data(True)  # FIXME: success value?


@asyncio.coroutine
def set_status(request):
    obj_type = request.match_info['type']
    obj_id = request.match_info['id']
    status = yield from decode_request(request)
    request.app['backend'].set_status(obj_type, obj_id, status)
    return encode_data(True)  # FIXME: success value?


@asyncio.coroutine
def put_bundle(request):
    obj_type = request.match_info['type']
    obj_id = request.match_info['id']

    # TODO: handle streaming properly
    content = yield from decode_request(request)
    request.app['backend'].cache.add(obj_type, obj_id, content)
    return encode_data(True)  # FIXME: success value?


@asyncio.coroutine
def send_notif(request):
    obj_type = request.match_info['type']
    obj_id = request.match_info['id']
    request.app['backend'].send_all_notifications(obj_type, obj_id)
    return encode_data(True)  # FIXME: success value?


# Batch endpoints

@asyncio.coroutine
def batch_cook(request):
    batch = yield from decode_request(request)
    for obj_type, obj_id in batch:
        if obj_type not in COOKER_TYPES:
            raise aiohttp.web.HTTPNotFound
    batch_id = request.app['backend'].batch_cook(batch)
    return encode_data({'id': batch_id})


@asyncio.coroutine
def batch_progress(request):
    batch_id = request.match_info['batch_id']
    bundles = request.app['backend'].batch_info(batch_id)
    if not bundles:
        raise aiohttp.web.HTTPNotFound
    bundles = [user_info(bundle) for bundle in bundles]
    counter = collections.Counter(b['status'] for b in bundles)
    res = {'bundles': bundles, 'total': len(bundles),
           **{k: 0 for k in ('new', 'pending', 'done', 'failed')},
           **dict(counter)}
    return encode_data(res)


# Web server

def make_app(backend, **kwargs):
    app = SWHRemoteAPI(**kwargs)
    app.router.add_route('GET', '/', index)

    # Endpoints used by the web API
    app.router.add_route('GET', '/fetch/{type}/{id}', vault_fetch)
    app.router.add_route('POST', '/cook/{type}/{id}', vault_cook)
    app.router.add_route('GET', '/progress/{type}/{id}', vault_progress)

    # Endpoints used by the Cookers
    app.router.add_route('POST', '/set_progress/{type}/{id}', set_progress)
    app.router.add_route('POST', '/set_status/{type}/{id}', set_status)
    app.router.add_route('POST', '/put_bundle/{type}/{id}', put_bundle)
    app.router.add_route('POST', '/send_notif/{type}/{id}', send_notif)

    # Endpoints for batch requests
    app.router.add_route('POST', '/batch_cook', batch_cook)
    app.router.add_route('GET', '/batch_progress/{batch_id}', batch_progress)

    app['backend'] = backend
    return app


def get_local_backend(config_file):
    if os.path.isfile(config_file):
        cfg = config.read(config_file, DEFAULT_CONFIG)
    else:
        cfg = config.load_named_config(config_file, DEFAULT_CONFIG)

    if 'vault' not in cfg:
        raise ValueError("missing '%vault' configuration")

    vcfg = cfg['vault']
    if vcfg['cls'] != 'local':
        raise EnvironmentError(
            "The vault backend can only be started with a 'local' "
            "configuration", err=True)
    args = vcfg['args']
    if 'cache' not in args:
        args['cache'] = cfg.get('cache')
    if 'storage' not in args:
        args['storage'] = cfg.get('storage')
    if 'scheduler' not in args:
        args['scheduler'] = cfg.get('scheduler')

    for key in ('cache', 'storage', 'scheduler'):
        if not args.get(key):
            raise ValueError(
                "invalid configuration; missing %s config entry." % key)

    return get_vault('local', args)


def make_app_from_configfile(config_file=DEFAULT_CONFIG_PATH, **kwargs):
    config_file = os.environ.get('SWH_CONFIG_FILENAME', config_file)
    vault = get_local_backend(config_file)
    return make_app(backend=vault, **kwargs)


if __name__ == '__main__':
    print('Deprecated. Use swh-vault ')



# Copyright (C) 2016-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import abc
import io
import logging

from swh.model import hashutil

MAX_BUNDLE_SIZE = 2 ** 29  # 512 MiB
DEFAULT_CONFIG_PATH = 'vault/cooker'
DEFAULT_CONFIG = {
    'storage': ('dict', {
        'cls': 'remote',
        'args': {
            'url': 'http://localhost:5002/',
        },
    }),
    'vault': ('dict', {
        'cls': 'remote',
        'args': {
            'url': 'http://localhost:5005/',
        },
    }),
    'max_bundle_size': ('int', MAX_BUNDLE_SIZE),
}


class PolicyError(Exception):
    """Raised when the bundle violates the cooking policy."""
    pass


class BundleTooLargeError(PolicyError):
    """Raised when the bundle is too large to be cooked."""
    pass


class BytesIOBundleSizeLimit(io.BytesIO):
    def __init__(self, *args, size_limit=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.size_limit = size_limit

    def write(self, chunk):
        if ((self.size_limit is not None
             and self.getbuffer().nbytes + len(chunk) > self.size_limit)):
            raise BundleTooLargeError(
                "The requested bundle exceeds the maximum allowed "
                "size of {} bytes.".format(self.size_limit))
        return super().write(chunk)


class BaseVaultCooker(metaclass=abc.ABCMeta):
    """Abstract base class for the vault's bundle creators

    This class describes a common API for the cookers.

    To define a new cooker, inherit from this class and override:
    - CACHE_TYPE_KEY: key to use for the bundle to reference in cache
    - def cook(): cook the object into a bundle
    """
    CACHE_TYPE_KEY = None

    def __init__(self, obj_type, obj_id, backend, storage,
                 max_bundle_size=MAX_BUNDLE_SIZE):
        """Initialize the cooker.

        The type of the object represented by the id depends on the
        concrete class. Very likely, each type of bundle will have its
        own cooker class.

        Args:
            obj_type: type of the object to be cooked into a bundle (directory,
                      revision_flat or revision_gitfast; see
                      swh.vault.cooker.COOKER_TYPES).
            obj_id: id of the object to be cooked into a bundle.
            backend: the vault backend (swh.vault.backend.VaultBackend).
        """
        self.obj_type = obj_type
        self.obj_id = hashutil.hash_to_bytes(obj_id)
        self.backend = backend
        self.storage = storage
        self.max_bundle_size = max_bundle_size

    @abc.abstractmethod
    def check_exists(self):
        """Checks that the requested object exists and can be cooked.

        Override this in the cooker implementation.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def prepare_bundle(self):
        """Implementation of the cooker. Yields chunks of the bundle bytes.

        Override this with the cooker implementation.
        """
        raise NotImplementedError

    def write(self, chunk):
        self.fileobj.write(chunk)

    def cook(self):
        """Cook the requested object into a bundle
        """
        self.backend.set_status(self.obj_type, self.obj_id, 'pending')
        self.backend.set_progress(self.obj_type, self.obj_id, 'Processing...')

        self.fileobj = BytesIOBundleSizeLimit(size_limit=self.max_bundle_size)
        try:
            self.prepare_bundle()
            bundle = self.fileobj.getvalue()
            # TODO: use proper content streaming instead of put_bundle()
            self.backend.put_bundle(self.CACHE_TYPE_KEY, self.obj_id, bundle)
        except PolicyError as e:
            self.backend.set_status(self.obj_type, self.obj_id, 'failed')
            self.backend.set_progress(self.obj_type, self.obj_id, str(e))
        except Exception:
            self.backend.set_status(self.obj_type, self.obj_id, 'failed')
            self.backend.set_progress(
                self.obj_type, self.obj_id,
                "Internal Server Error. This incident will be reported.")
            logging.exception("Bundle cooking failed.")
        else:
            self.backend.set_status(self.obj_type, self.obj_id, 'done')
            self.backend.set_progress(self.obj_type, self.obj_id, None)
        finally:
            self.backend.send_notif(self.obj_type, self.obj_id)



# Copyright (C) 2016-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.model import hashutil
from swh.core.api import SWHRemoteAPI


class VaultAPIError(Exception):
    """Vault API Error"""
    def __str__(self):
        return ('An unexpected error occurred in the Vault backend: {}'
                .format(self.args))


class RemoteVaultClient(SWHRemoteAPI):
    """Client to the Software Heritage vault cache."""

    def __init__(self, url, timeout=None):
        super().__init__(
            api_exception=VaultAPIError, url=url, timeout=timeout)

    # Web API endpoints

    def fetch(self, obj_type, obj_id):
        hex_id = hashutil.hash_to_hex(obj_id)
        return self.get('fetch/{}/{}'.format(obj_type, hex_id))

    def cook(self, obj_type, obj_id, email=None):
        hex_id = hashutil.hash_to_hex(obj_id)
        return self.post('cook/{}/{}'.format(obj_type, hex_id),
                         data={},
                         params=({'email': email} if email else None))

    def progress(self, obj_type, obj_id):
        hex_id = hashutil.hash_to_hex(obj_id)
        return self.get('progress/{}/{}'.format(obj_type, hex_id))

    # Cookers endpoints

    def set_progress(self, obj_type, obj_id, progress):
        hex_id = hashutil.hash_to_hex(obj_id)
        return self.post('set_progress/{}/{}'.format(obj_type, hex_id),
                         data=progress)

    def set_status(self, obj_type, obj_id, status):
        hex_id = hashutil.hash_to_hex(obj_id)
        return self.post('set_status/{}/{}' .format(obj_type, hex_id),
                         data=status)

    # TODO: handle streaming properly
    def put_bundle(self, obj_type, obj_id, bundle):
        hex_id = hashutil.hash_to_hex(obj_id)
        return self.post('put_bundle/{}/{}' .format(obj_type, hex_id),
                         data=bundle)

    def send_notif(self, obj_type, obj_id):
        hex_id = hashutil.hash_to_hex(obj_id)
        return self.post('send_notif/{}/{}' .format(obj_type, hex_id),
                         data=None)

    # Batch endpoints

    def batch_cook(self, batch):
        return self.post('batch_cook', data=batch)

    def batch_progress(self, batch_id):
        return self.get('batch_progress/{}'.format(batch_id))



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from unittest.mock import MagicMock

from swh.model import hashutil
from swh.vault.cookers.base import BaseVaultCooker


TEST_BUNDLE_CHUNKS = [b"test content 1\n",
                      b"test content 2\n",
                      b"test content 3\n"]
TEST_BUNDLE_CONTENT = b''.join(TEST_BUNDLE_CHUNKS)
TEST_OBJ_TYPE = 'test_type'
TEST_HEX_ID = '17a3e48bce37be5226490e750202ad3a9a1a3fe9'
TEST_OBJ_ID = hashutil.hash_to_bytes(TEST_HEX_ID)


class BaseVaultCookerMock(BaseVaultCooker):
    CACHE_TYPE_KEY = TEST_OBJ_TYPE

    def __init__(self):
        # we do not call super() here to bypass the building of db objects from
        # config since we do mock these db objects
        self.config = {}
        self.storage = MagicMock()
        self.backend = MagicMock()
        self.obj_type = self.CACHE_TYPE_KEY
        self.obj_id = hashutil.hash_to_bytes(TEST_OBJ_ID)
        self.max_bundle_size = 1024

    def check_exists(self):
        return True

    def prepare_bundle(self):
        for chunk in TEST_BUNDLE_CHUNKS:
            self.write(chunk)


def test_simple_cook():
    cooker = BaseVaultCookerMock()
    cooker.cook()
    cooker.backend.put_bundle.assert_called_once_with(
        TEST_OBJ_TYPE, TEST_OBJ_ID, TEST_BUNDLE_CONTENT)
    cooker.backend.set_status.assert_called_with(
        TEST_OBJ_TYPE, TEST_OBJ_ID, 'done')
    cooker.backend.set_progress.assert_called_with(
        TEST_OBJ_TYPE, TEST_OBJ_ID, None)
    cooker.backend.send_notif.assert_called_with(
        TEST_OBJ_TYPE, TEST_OBJ_ID)


def test_code_exception_cook():
    cooker = BaseVaultCookerMock()
    cooker.prepare_bundle = MagicMock()
    cooker.prepare_bundle.side_effect = RuntimeError("Nope")
    cooker.cook()

    # Potentially remove this when we have objstorage streaming
    cooker.backend.put_bundle.assert_not_called()

    cooker.backend.set_status.assert_called_with(
        TEST_OBJ_TYPE, TEST_OBJ_ID, 'failed')
    assert "Nope" not in cooker.backend.set_progress.call_args[0][2]
    cooker.backend.send_notif.assert_called_with(TEST_OBJ_TYPE, TEST_OBJ_ID)


def test_policy_exception_cook():
    cooker = BaseVaultCookerMock()
    cooker.max_bundle_size = 8
    cooker.cook()

    # Potentially remove this when we have objstorage streaming
    cooker.backend.put_bundle.assert_not_called()

    cooker.backend.set_status.assert_called_with(
        TEST_OBJ_TYPE, TEST_OBJ_ID, 'failed')
    assert "exceeds" in cooker.backend.set_progress.call_args[0][2]
    cooker.backend.send_notif.assert_called_with(TEST_OBJ_TYPE, TEST_OBJ_ID)



# Copyright (C) 2017  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import functools
import os
import time
import zlib

from fastimport.commands import (CommitCommand, ResetCommand, BlobCommand,
                                 FileDeleteCommand, FileModifyCommand)

from swh.model import hashutil
from swh.model.toposort import toposort
from swh.model.from_disk import mode_to_perms
from swh.vault.cookers.base import BaseVaultCooker
from swh.vault.to_disk import get_filtered_files_content


class RevisionGitfastCooker(BaseVaultCooker):
    """Cooker to create a git fast-import bundle """
    CACHE_TYPE_KEY = 'revision_gitfast'

    def check_exists(self):
        return not list(self.storage.revision_missing([self.obj_id]))

    def prepare_bundle(self):
        self.log = list(toposort(self.storage.revision_log([self.obj_id])))
        self.gzobj = zlib.compressobj(9, zlib.DEFLATED, zlib.MAX_WBITS | 16)
        self.fastexport()
        self.write(self.gzobj.flush())

    def write_cmd(self, cmd):
        chunk = bytes(cmd) + b'\n'
        super().write(self.gzobj.compress(chunk))

    def fastexport(self):
        """Generate all the git fast-import commands from a given log.
        """
        self.rev_by_id = {r['id']: r for r in self.log}
        self.obj_done = set()
        self.obj_to_mark = {}
        self.next_available_mark = 1

        last_progress_report = None

        for i, rev in enumerate(self.log, 1):
            # Update progress if needed
            ct = time.time()
            if (last_progress_report is None
                    or last_progress_report + 2 <= ct):
                last_progress_report = ct
                pg = ('Computing revision {}/{}'.format(i, len(self.log)))
                self.backend.set_progress(self.obj_type, self.obj_id, pg)

            # Compute the current commit
            self._compute_commit_command(rev)

    def mark(self, obj_id):
        """Get the mark ID as bytes of a git object.

        If the object has not yet been marked, assign a new ID and add it to
        the mark dictionary.
        """
        if obj_id not in self.obj_to_mark:
            self.obj_to_mark[obj_id] = self.next_available_mark
            self.next_available_mark += 1
        return str(self.obj_to_mark[obj_id]).encode()

    def _compute_blob_command_content(self, file_data):
        """Compute the blob command of a file entry if it has not been
        computed yet.
        """
        obj_id = file_data['sha1']
        if obj_id in self.obj_done:
            return
        contents = list(get_filtered_files_content(self.storage, [file_data]))
        content = contents[0]['content']
        self.write_cmd(BlobCommand(mark=self.mark(obj_id), data=content))
        self.obj_done.add(obj_id)

    def _author_tuple_format(self, author, date):
        # We never want to have None values here so we replace null entries
        # by ''.
        if author is not None:
            author_tuple = (author.get('name') or b'',
                            author.get('email') or b'')
        else:
            author_tuple = (b'', b'')
        if date is not None:
            date_tuple = (date.get('timestamp', {}).get('seconds') or 0,
                          (date.get('offset') or 0) * 60)
        else:
            date_tuple = (0, 0)
        return author_tuple + date_tuple

    def _compute_commit_command(self, rev):
        """Compute a commit command from a specific revision.
        """
        if 'parents' in rev and rev['parents']:
            from_ = b':' + self.mark(rev['parents'][0])
            merges = [b':' + self.mark(r) for r in rev['parents'][1:]]
            parent = self.rev_by_id[rev['parents'][0]]
        else:
            # We issue a reset command before all the new roots so that they
            # are not automatically added as children of the current branch.
            self.write_cmd(ResetCommand(b'refs/heads/master', None))
            from_ = None
            merges = None
            parent = None

        # Retrieve the file commands while yielding new blob commands if
        # needed.
        files = list(self._compute_file_commands(rev, parent))

        # Construct and write the commit command
        author = self._author_tuple_format(rev['author'], rev['date'])
        committer = self._author_tuple_format(rev['committer'],
                                              rev['committer_date'])
        self.write_cmd(CommitCommand(
            ref=b'refs/heads/master',
            mark=self.mark(rev['id']),
            author=author,
            committer=committer,
            message=rev['message'] or b'',
            from_=from_,
            merges=merges,
            file_iter=files))

    @functools.lru_cache(maxsize=4096)
    def _get_dir_ents(self, dir_id=None):
        """Get the entities of a directory as a dictionary (name -> entity).

        This function has a cache to avoid doing multiple requests to retrieve
        the same entities, as doing a directory_ls() is expensive.
        """
        data = (self.storage.directory_ls(dir_id)
                if dir_id is not None else [])
        return {f['name']: f for f in data}

    def _compute_file_commands(self, rev, parent=None):
        """Compute all the file commands of a revision.

        Generate a diff of the files between the revision and its main parent
        to find the necessary file commands to apply.
        """
        # Initialize the stack with the root of the tree.
        cur_dir = rev['directory']
        parent_dir = parent['directory'] if parent else None
        stack = [(b'', cur_dir, parent_dir)]

        while stack:
            # Retrieve the current directory and the directory of the parent
            # commit in order to compute the diff of the trees.
            root, cur_dir_id, prev_dir_id = stack.pop()
            cur_dir = self._get_dir_ents(cur_dir_id)
            prev_dir = self._get_dir_ents(prev_dir_id)

            # Find subtrees to delete:
            #  - Subtrees that are not in the new tree (file or directory
            #    deleted).
            #  - Subtrees that do not have the same type in the new tree
            #    (file -> directory or directory -> file)
            # After this step, every node remaining in the previous directory
            # has the same type than the one in the current directory.
            for fname, f in prev_dir.items():
                if ((fname not in cur_dir
                     or f['type'] != cur_dir[fname]['type'])):
                    yield FileDeleteCommand(path=os.path.join(root, fname))

            # Find subtrees to modify:
            #  - Leaves (files) will be added or modified using `filemodify`
            #  - Other subtrees (directories) will be added to the stack and
            #    processed in the next iteration.
            for fname, f in cur_dir.items():
                # A file is added or modified if it was not in the tree, if its
                # permissions changed or if its content changed.
                if (f['type'] == 'file'
                    and (fname not in prev_dir
                         or f['sha1'] != prev_dir[fname]['sha1']
                         or f['perms'] != prev_dir[fname]['perms'])):
                    # Issue a blob command for the new blobs if needed.
                    self._compute_blob_command_content(f)
                    yield FileModifyCommand(
                        path=os.path.join(root, fname),
                        mode=mode_to_perms(f['perms']).value,
                        dataref=(b':' + self.mark(f['sha1'])),
                        data=None)
                # A revision is added or modified if it was not in the tree or
                # if its target changed
                elif (f['type'] == 'rev'
                      and (fname not in prev_dir
                           or f['target'] != prev_dir[fname]['target'])):
                    yield FileModifyCommand(
                        path=os.path.join(root, fname),
                        mode=0o160000,
                        dataref=hashutil.hash_to_hex(f['target']).encode(),
                        data=None)
                # A directory is added or modified if it was not in the tree or
                # if its target changed.
                elif f['type'] == 'dir':
                    f_prev_target = None
                    if fname in prev_dir and prev_dir[fname]['type'] == 'dir':
                        f_prev_target = prev_dir[fname]['target']
                    if f_prev_target is None or f['target'] != f_prev_target:
                        stack.append((os.path.join(root, fname),
                                      f['target'], f_prev_target))
